#include <owl/pch.h>
#include <owl/applicat.h>
#include <mmsystem.h>
#include <windows.h>
#include <math.h>

#define BUF_SIZE   2048
#define FREQ       20480
#define LEVELS     20
#define LEV_HEIGHT 5
#define LEV_WIDTH  6
#define FREQS      74

const double INC = 1.0 / LEVELS;
const double M1DSQR2PI = 1.0 / sqrt(M_PI * 2);

WAVEINCAPS wavecaps;
unsigned char buffer[BUF_SIZE];
char temp[10];
//unsigned char pixbuf[BUF_SIZE / 4];
double freqs[FREQS];
double re[BUF_SIZE];
double im[BUF_SIZE];
double amp[BUF_SIZE / 2];

class TMainWindow: public TWindow {
    HWAVEIN h;
    WAVEHDR buf;
    WAVEFORMATEX wavefmt;
    void Init();
  public:
    TMainWindow();
    void SetupWindow();
         LRESULT DataReady(unsigned int, long);
    void Paint(TDC&, bool, TRect&);
    void DrawFFT(HDC, int, int);
    void PerformFFT(int);
    ~TMainWindow()
    {
      waveInClose(h);
    }
  DECLARE_RESPONSE_TABLE(TMainWindow);
};

DEFINE_RESPONSE_TABLE1(TMainWindow, TWindow)
  EV_MESSAGE(MM_WIM_DATA, DataReady),
  EV_WM_PAINT,
END_RESPONSE_TABLE;

TMainWindow::TMainWindow(): TWindow(0, 0, 0)
{
}

void TMainWindow::Init()
{
  for (int i = 0; i < FREQS; i++)
    freqs[i] = 0.0;

  waveInClose(h);

  wavefmt.wFormatTag = WAVE_FORMAT_PCM;
  wavefmt.nChannels = 1;
  wavefmt.nSamplesPerSec = FREQ;
  wavefmt.nAvgBytesPerSec = BUF_SIZE;
  wavefmt.nBlockAlign = 2;
  wavefmt.wBitsPerSample = 8;
  wavefmt.cbSize = 0;

  MMRESULT err = waveInOpen(&h, 0, &wavefmt, (DWORD)HWindow, 0, CALLBACK_WINDOW);
  char errstr[100] = "Error opening waveform-audio input device!\nerror ";

  if (err != MMSYSERR_NOERROR)
  {
    ltoa(err, temp, 10);
    strcat(errstr, temp);
    MessageBox(errstr, "Error", MB_OK | MB_ICONSTOP);
  }
}

void TMainWindow::SetupWindow()
{
  TWindow::SetupWindow();
  SetBkgndColor(TColor::Black);

  Init();

  buf.lpData = buffer;
  buf.dwBufferLength = BUF_SIZE;
  buf.dwUser = 0;

  waveInPrepareHeader(h, &buf, sizeof(buf));
  waveInAddBuffer(h, &buf, sizeof(buf));
  waveInStart(h);
}

LRESULT TMainWindow::DataReady(unsigned int wParam, long lParam)
{
  waveInPrepareHeader(h, &buf, sizeof(buf));
  waveInAddBuffer(h, &buf, sizeof(buf));
  PerformFFT(BUF_SIZE);
  InvalidateRect(GetClientRect(), FALSE);
  return 0;
}

void TMainWindow::PerformFFT(int N)
{
  int NM1 = N - 1;
  int ND2 = N >> 1;
  int M = int(log(N) / log(2));
  int j = ND2, i = 0, l, k, le, le2, ip, jm1;
  double tr, ti, ur, ui, sr, si, max = 0, max2 = 0, t, s;

  for (; i < N; i++)
  {
    re[i] = double(buffer[i] - 128) / 128.0;
    t = fabs(re[i]);
    if (t > max2) max2 = t;
    s += re[i];
    im[i] = 0.0;
  }
  max2 -= s / N;

  for (i = 1; i <= N - 2; i++)
  {
    if (i < j)
    {
      tr = re[j];
      ti = im[j];
      re[j] = re[i];
      im[j] = im[i];
      re[i] = tr;
      im[i] = ti;
    }
    k = ND2;
    while (k <= j)
    {
      j -= k;
      k >>= 1;
    }
    j += k;
  }
  for (l = 1; l <= M; l++)
  {
    le = 1 << l;
    le2 = le >> 1;
    ur = 1;
    ui = 0;
    sr = cos(M_PI / le2);
    si = -sin(M_PI / le2);
    for (j = 1; j <= le2; j++)
    {
      jm1 = j - 1;
      for (i = jm1; i <= NM1; i += le)
      {
        ip = i + le2;
        tr = re[ip] * ur - im[ip] * ui;
        ti = re[ip] * ui + im[ip] * ur;
        re[ip] = re[i] - tr;
        im[ip] = im[i] - ti;
        re[i] += tr;
        im[i] += ti;
      }
      tr = ur;
      ur = tr * sr - ui * si;
      ui = tr * si + ui * sr;
    }
  }

  j = 0;

  for (i = 2; i < 10; i++, j++)
  {
    freqs[j] = sqrt(re[i] * re[i] + im[i] * im[i]);
    if (freqs[j] > max) max = freqs[j];
  }

  for (i = 10; i < 100; i += 2, j++)
  {
    freqs[j] = sqrt(re[i] * re[i] + im[i] * im[i]);
    if (freqs[j] > max) max = freqs[j];
  }

  for (i = 100; i <= 300; i += 10, j++)
  {
    freqs[j] = sqrt(re[i] * re[i] + im[i] * im[i]);
    if (freqs[j] > max) max = freqs[j];
  }

  if (max > 0 && max2 > 0)
  {
    if (max < INC) max = INC;
    max = max2 / max;
    for (i = 0; i < FREQS; i++)
      freqs[i] *= max;
  }
}

void TMainWindow::Paint(TDC& dc, bool erase, TRect& rect)
{
  TRect r;
  HDC MemDC;
  HBITMAP Bitmap;

  dc.SelectObject((HPEN)GetStockObject(BLACK_PEN));
  dc.SelectObject((HBRUSH)GetStockObject(BLACK_BRUSH));
  dc.Rectangle(0, 0, BUF_SIZE / 4, 128);

  dc.MoveTo(0, (unsigned char)buf.lpData[0] >> 1);
  dc.SelectObject((HPEN)GetStockObject(WHITE_PEN));
  for (int i = 1; i < Attr.W; i++)
    dc.LineTo(i, (unsigned char)buf.lpData[i] >> 1);

  MemDC = CreateCompatibleDC(dc.GetHDC());
  Bitmap = this -> GetModule() -> LoadBitmap(1);
  SelectObject(MemDC, Bitmap);
  BitBlt(dc.GetHDC(), 26, 242, 450, 28, MemDC, 0, 0, SRCCOPY);
  Bitmap = this -> GetModule() -> LoadBitmap(2);
  DeleteObject((HBITMAP)SelectObject(MemDC, Bitmap));
  BitBlt(dc.GetHDC(), 8, 135, 20, 108, MemDC, 0, 0, SRCCOPY);
  DeleteDC(MemDC);
  DeleteObject(Bitmap);

  DrawFFT(dc.GetHDC(), 30, 240);
}

void TMainWindow::DrawFFT(HDC dc, int x, int y)
{
  double d, v, C = 255.0 / LEVELS;
  int i = 0, j;
  LOGBRUSH br;
  HBRUSH brush[LEVELS], old_brush;
  HPEN pen[LEVELS], old_pen;
  br.lbStyle = BS_SOLID;
  br.lbHatch = 0;

  for (; i < LEVELS; i++)
  {
    br.lbColor = RGB(255, 255 - int(i * C), 128 - (int(i * C) >> 1));
    brush[i] = CreateBrushIndirect(&br);
    pen[i] = CreatePen(PS_SOLID, 1, RGB(255, 255 - int(i * C), 128 - (int(i * C) >> 1)));
  }

  old_brush = (HBRUSH)SelectObject(dc, GetStockObject(BLACK_BRUSH));
  old_pen = (HPEN)SelectObject(dc, GetStockObject(BLACK_PEN));

  for (i = 0; i < FREQS; i++)
  {
    v = freqs[i];
    d = 0.0;
    for (j = 0; j < LEVELS; j++)
    {
      if (v > d)
      {
        SelectObject(dc, brush[j]);
        SelectObject(dc, pen[j]);
      }
      else
      {
        SelectObject(dc, GetStockObject(BLACK_BRUSH));
        SelectObject(dc, GetStockObject(BLACK_PEN));
      }
      Rectangle(dc, x + i * LEV_WIDTH, y - j * LEV_HEIGHT, x + (i + 1) * LEV_WIDTH - 1, y - (j + 1) * LEV_HEIGHT + 1);
      d += INC;
    }
  }

  SelectObject(dc, old_brush);
  SelectObject(dc, old_pen);

  for (i = 0; i < LEVELS; i++)
  {
    DeleteObject(pen[i]);
    DeleteObject(brush[i]);
  }
}

class TMainApp : public TApplication {
  public:
    TMainApp() : TApplication() {}
    void InitMainWindow();
};

void TMainApp::InitMainWindow()
{
  TFrameWindow *W = new TFrameWindow(0, "The Fast Fourier Transform - by HackMaster", new TMainWindow);
  W -> Attr.W = BUF_SIZE / 4;
  W -> Attr.H = 300;
  W -> Attr.Style = WS_DLGFRAME | WS_SYSMENU | WS_MINIMIZEBOX;
  MainWindow = W;
}

int OwlMain(int, char* [])
{
  if (!waveInGetNumDevs)
  {
    MessageBox(0, "No waveform-audio input devices!", "Error", MB_OK + MB_ICONSTOP);
    return 1;
  }
  return TMainApp().Run();
}

